# -*- coding: utf-8 -*-
'''
        uml-tool
        plantuml (www.plantuml.com) based application to simplify documentation
        (c) 2016 Melf Pahlke
        
        19.07.16 - MainFrame
        20.07.16 - MainFrame.Notebook

        uses:
        Python> 2.7.9            www.python.org
                - easy_install  https://pypi.python.org/pypi/setuptools (https://bootstrap.pypa.io/ez_setup.py)
                        - run script to install package
                - open command line
                        - run command "easy_install Pillow"
                        - close all Python Applications (IDLE, ...)
                        
        PlantUML                www.plantuml.com/download.html (as executable .jar file)
        Graphviz >= 2.26.3      www.graphviz.org (http://www.graphviz.org/pub/graphviz/stable/windows/graphviz-2.38.msi or other)
        
'''
try:
        import tkinter as tk
        import ttk
        import tkMessageBox
        import tkFileDialog 
        from PIL import Image, ImageTk #needs to run "easy_install Pillow" in command line
        #import Image
except ImportError:
        import Tkinter as tk
        import ttk
        import tkMessageBox
        import tkFileDialog 
        from PIL import Image, ImageTk #needs to run "easy_install Pillow" in command line
        #import Image
        
import sys
from subprocess import call,CalledProcessError
from ast import literal_eval
from time import sleep

#if sys.partition.startswith('linux2') or sys.partition.startswith('win32') or sys.partition.startswith('win64'):
from os.path import isfile

class MainFrame():
        def __init__(self,defaults):
                self.__root = tk.Tk()
                self.__defaults = defaults
                self.__main_menu_commands = None
                self.__menu = self.__create_menu()
                self.__root.wm_title(str(defaults['mainTitle']))
                self.__root.geometry(str(defaults['mainWidth'])+'x'+str(defaults['mainHeight'])+'+'+str(defaults['mainXOffset'])+'+'+str(defaults['mainYOffset']))
                self.__nb = None
                self.__tabs = None
                self.__texts = None
                self.__undoHistory = None
                self.__root.config(menu=self.__menu)
                self.__root.protocol("WM_DELETE_WINDOW",self.__quit_click)
                self.__root.bind_all("<Control-n>",self.__new_click)
                self.__root.bind_all("<Control-s>",self.__save_click)
                self.__root.bind_all("<Control-Shift-S>",self.__save_as_click)
                self.__root.bind_all("<Control-l>",self.__load_click)
                self.__root.bind_all("<Control-z>",self.__undo_click)
                self.__root.bind_all("<Control-a>",self.__select_all_click)
                #self.__root.bind_all("<Control-c>",self.__copy_click)
                #self.__root.bind_all("<Control-v>",self.__paste_click)
                self.__root.bind_all("<Control-x>",self.__cut_click)
                self.__root.bind("<Configure>",self.__changed_size)
                #### ADD file location below if you want another project structure ###
                self.__root.iconbitmap(r"icons/plantuml.ico")
                self.__root.mainloop()

        def __create_menu(self):
                #print 'create_menu()'
                m = tk.Menu(self.__root)
                #FILE
                fileMenu = tk.Menu(m, tearoff=0)
                fileMenu.add_command(label="New", command=lambda:self.__new_click(fileMenu),accelerator="Ctrl+N")
                fileMenu.add_command(label="Close current", command=lambda:self.__close_click(fileMenu),state="disabled")
                fileMenu.add_separator()
                fileMenu.add_command(label="Save", command=lambda:self.__save_click(fileMenu),state="disabled",accelerator="Ctrl+S")
                fileMenu.add_command(label="Save As ...", command=lambda:self.__save_as_click(fileMenu),state="disabled",accelerator="Ctrl+Shift+S")
                fileMenu.add_command(label="Load", command=lambda:self.__load_click(fileMenu),accelerator="Ctrl+L")
                fileMenu.add_separator()
                fileMenu.add_command(label="Quit", command=self.__quit_click)
                #EDIT
                editMenu = tk.Menu(m, tearoff=0)
                editMenu.add_command(label="Undo",command=lambda:self.__undo_click(editMenu),state=tk.DISABLED,accelerator="Ctrl+Z")
                editMenu.add_separator()
                editMenu.add_command(label="Select All",command=lambda:self.__select_all_click(editMenu),accelerator="Ctrl+A")
                editMenu.add_command(label="Copy (use \"Ctrl+S\")",command=lambda:self.__copy_click(editMenu),state="disabled",accelerator="Ctrl+C")
                editMenu.add_command(label="Paste (use \"Ctrl+V\")",command=lambda:self.__paste_click(editMenu),state="disabled",accelerator="Ctrl+V")
                editMenu.add_command(label="Cut (use \"Ctrl+X\")",command=lambda:self.__cut_click(editMenu),state="disabled",accelerator="Ctrl+X")
                #VIEW
                viewMenu = tk.Menu(m, tearoff=0)
                #UML
                #img = Image.open(
                #playImage = ImageTk.PhotoImage(img)

                #### ADD file location below if you want another project structure ###
                
                playImage = tk.PhotoImage(file=r"icons/play.gif")
                umlMenu = tk.Menu(m, tearoff=0)
                s = None
                umlMenu.add_command(label="Play",image=playImage,command=lambda:self.__play_click(umlMenu),accelerator="F5")
                self.__play_image = playImage
                #OTHER
                otherMenu = tk.Menu(m, tearoff=0)
                otherMenu.add_command(label="Preferences",command=lambda:self.__preferences_click(otherMenu),state=tk.DISABLED)
                otherMenu.add_separator()
                otherMenu.add_command(label="About",command=self.__about_click)
                #Assemble Main-Menu
                self.__main_menu_commands = {}
                m.add_cascade(label="File",menu=fileMenu)
                self.__main_menu_commands['File'] = fileMenu
                m.add_cascade(label="Edit",menu=editMenu)
                self.__main_menu_commands['Edit'] = editMenu
                m.add_cascade(label="View",menu=viewMenu)
                self.__main_menu_commands['View'] = viewMenu
                m.add_cascade(label="UML",menu=umlMenu)
                self.__main_menu_commands['UML'] = umlMenu
                m.add_cascade(label="Other",menu=otherMenu)
                self.__main_menu_commands['Other'] = otherMenu
                return m

        def __get_current_selected_tab(self):
                #returns (index,text)
                i = self.__nb.select()
                if i != None:
                        return (i, self.__nb.tab(i,'text'))
                else:
                        return None
                        
        def __get_current_selection(self):
                pass
        
        def __new_click(self,w=None):
                #print "__new_click()"
                if self.__nb == None:
                        self.__nb = ttk.Notebook(self.__root)
                        self.__nb.pack(fill='both',expand=1)
                if self.__tabs == None:
                        self.__tabs = {}
                        self.__texts = {}
                new_tab = ttk.Frame(self.__nb)
                new_text = tk.Text(new_tab)
                new_scrollbar = tk.Scrollbar(new_tab,command=new_text.yview)
                new_text.config(yscrollcommand=new_scrollbar.set)
                i = 1
                while 1:
                        if not 'untitled'+str(i)+'*' in self.__tabs:
                                self.__nb.add(new_tab, text="untitled"+str(i)+'*',compound=tk.TOP)
                                self.__tabs['untitled'+str(i)+'*'] = new_tab
                                self.__texts['untitled'+str(i)+'*'] = [new_text, new_scrollbar, '']
                                new_scrollbar.pack(side='right',fill='y',expand=0)
                                new_text.pack(side='left',fill='both',expand=1)
                                new_text.bind("<KeyRelease>",self.__key_released_in_text)
                                new_text.bind("<KeyPress>",self.__key_pressed_in_text)
                                self.__undoHistory = {"untitled"+str(i)+'*':['']}
                                #print self.__tabs,self.__texts
                                break
                        else:
                                i +=1
                #activate 
                w.entryconfig(1,state="normal")#"close current" button
                w.entryconfig(3,state="normal")#"save" button
                w.entryconfig(4,state="normal")#"save as ..." button
                return

        def __close_click(self,w=None):
                #get current selected tab
                t = self.__get_current_selected_tab()
                if t != None:
                        #print t
                        if '*' in str(t[1]):
                                #unsaved
                                if tkMessageBox.askyesno("Save before closing?","Do you want to save before you close?"):
                                        #save
                                        self.__save_as_click(w)
                        #close
                        del self.__texts[str(t[1])]
                        self.__nb.forget(t[0]) #remove tab
                        del self.__tabs[str(t[1])]
                        if len(self.__tabs) == 0:
                                #nothing to close
                                w.entryconfig(1,state="disabled")
                                w.entryconfig(3,state="disabled")#"save" button (+1 due to separator in fileMenu)
                                w.entryconfig(4,state="disabled")#"save as ..." button
                else:
                        #nothing to close
                        w.entryconfig(1,state="disabled")
                        w.entryconfig(3,state="disabled")#"save" button
                        w.entryconfig(4,state="disabled")#"save as ..." button

        def __save_click(self,w=None):
                #print '__save_click()'
                #get current selected tab
                name = self.__get_current_selected_tab()[1]
                if name in self.__texts:
                        if self.__texts[name][2] != '':
                                fAddress = self.__texts[name][2]
                                #print fAddress
                                currText = self.__texts[name][0].get(1.0,tk.END)
                                try:
                                        with open(fAddress,'w') as f:
                                                f.write(currText)
                                        if self.__mark_as_saved(name) == 0:
                                                return 0
                                        else:
                                                return -1
                                except IOError:
                                        return 1
                        else:
                                #trigger "save as" routine
                                self.__save_as_click(w=w)
                                return
                                                
        
        def __save_as_click(self,w=None):
                #print '__save_as_click()'
                #get current selected tab
                name = self.__get_current_selected_tab()[1]
                #get filename
                options = {}
                options['defaultextension'] = '.txt'
                options['filetypes'] = [('text files', '.txt')]#('all files', '.*'),
                options['initialdir'] = self.__defaults['workingDirectory']
                fAddress = tkFileDialog.asksaveasfilename(**options)
                #print fAddress
                if fAddress != '' and fAddress != None:
                        #if isfile(fAddress):   #is already handled by tk
                        #        if not tkMessageBox.askyesno("Overwrite file?","The choosen file is existing.\nDo you want to overwrite it?"):
                        #                return -1
                        currText = self.__texts[name][0].get(1.0,tk.END)
                        try:
                                with open(fAddress,'w') as f:
                                        f.write(currText)
                                newName = fAddress[fAddress.rindex('/')+1:]
                                #print newName
                                #print fAddress
                                #change tab name to newName
                                tabBuf = None
                                textBuf = []
                                if name in self.__tabs:
                                        tabBuf = self.__tabs[name]
                                        del self.__tabs[name]
                                if name in self.__texts:
                                        textBuf = self.__texts[name]
                                        del self.__texts[name]
                                self.__tabs[newName] = tabBuf
                                self.__texts[newName] = textBuf
                                self.__texts[newName][2] = fAddress
                                self.__nb.tab(self.__nb.select(),text=newName)
                                #print self.__texts
                                #print self.__tabs
                        except IOError:
                                return 1
                        except IndexError:
                                return 2
        
        def __load_click(self,w=None):
                #print '__load_click()'
                #get file address
                options = {}
                options['defaultextension'] = '.txt'
                options['filetypes'] = [('all files', '.*'), ('text files', '.txt')]
                options['initialdir'] = self.__defaults['workingDirectory']
                #fAddress = tkFileDialog.askopenfilename(initialdir=options)
                fAddress = tkFileDialog.askopenfilename(**options)
                fName = fAddress[fAddress.rfind('/')+1:]
                #print fAddress
                #print fName
                if isfile(fAddress):
                        #create widgets
                        if self.__nb == None:
                                self.__nb = ttk.Notebook(self.__root)
                                self.__nb.pack(fill='both',expand=1)
                        if self.__tabs == None:
                                self.__tabs = {}
                                self.__texts = {}
                        new_tab = ttk.Frame(self.__nb)
                        new_text = tk.Text(new_tab)
                        new_scrollbar = tk.Scrollbar(new_tab,command=new_text.yview)
                        new_text.config(yscrollcommand=new_scrollbar.set)
                        i = 1
                        while 1:
                                if not str(fName) in self.__tabs:
                                        self.__nb.add(new_tab, text=str(fName),compound=tk.TOP)
                                        self.__tabs[str(fName)] = new_tab
                                        self.__texts[str(fName)] = (new_text, new_scrollbar, str(fAddress))
                                        new_scrollbar.pack(side='right',fill='y',expand=0)
                                        new_text.pack(side='left',fill='both',expand=1)
                                        new_text.bind("<KeyRelease>",self.__key_released_in_text)
                                        new_text.bind("<KeyPress>",self.__key_pressed_in_text)
                                        #print self.__tabs,self.__texts
                                        #fetch content from file
                                        content = ""
                                        with open(str(fAddress),'r') as f:
                                                content = f.read()
                                        #for l in lines:
                                        #        content = content+l
                                        new_text.delete(1.0, tk.END) #delete all text of text-widget
                                        new_text.insert(tk.END,content)
                                        self.__undoHistory = {str(fName)+'*':[content]}
                                        #print self.__undoHistory[str(fName)+'*']
                                        break
                                else:
                                        i +=1
                        #activate 
                        self.__main_menu_commands['File'].entryconfig(1,state="normal")#activate "close current" button
                        self.__main_menu_commands['File'].entryconfig(3,state="normal")#activate "save" button
                        self.__main_menu_commands['File'].entryconfig(4,state="normal")#activate "save as ..." button
                return
        
        def __quit_click(self):
                if tkMessageBox.askyesno("Quit "+self.__defaults['mainTitle'],"Do you want to quit the application?\nUnsaved changes will be lost."):
                        self.__root.destroy()
        
        def __undo_click(self,w=None):
                #print '__undo_click()'
                #get current tab widget name
                name = self.__get_current_selected_tab()[1]
                #print name
                #check undo_history
                if name in self.__undoHistory:
                        #print str(name)+' in self.__undoHistory'
                        if len(self.__undoHistory[name]) > 0:
                                #print 'Undo'
                                if len(self.__undoHistory[name]) > 1:
                                        self.__undoHistory[name].pop()
                                undoText = self.__undoHistory[name].pop()
                                #print undoText
                                #delete current Text
                                self.__texts[name][0].delete(1.0,tk.END)
                                self.__texts[name][0].insert(tk.END, undoText.encode('utf-8'))
                        if len(self.__undoHistory[name]) == 0:
                                w.entryconfig(0,state="disabled")
                return
        
        def __select_all_click(self,w=None):
                #get current tab widget name
                name = self.__get_current_selected_tab()[1]
                if name in self.__texts:
                        self.__texts[name][0].focus_set()
                        self.__texts[name][0].mark_set(tk.INSERT,1.0)
                        self.__texts[name][0].tag_add(tk.SEL,1.0,tk.END)
        
        def __copy_click(self,w=None):
                ##print '__copy_click()'
                #name = self.__get_current_selected_tab()[1]
                #if name in self.__texts:
                #        self.__root.clipboard_clear()
                #        self.__root.clipboard_append(self.__texts[name][0].selection_get())
                #        #print self.__root.selection_get(selection="CLIPBOARD")
                pass #is already integrated with tk
        
        def __paste_click(self,w=None):
                ##print '__paste_click()'
                #name = self.__get_current_selected_tab()[1]
                #if name in self.__texts:
                #        paste = self.__root.selection_get(selection="CLIPBOARD")
                #        #print paste
                #        self.__texts[name][0].insert(tk.CURRENT,str(paste))
                pass #is already integrated with tk
        
        def __cut_click(self,w=None):
                pass #is already integrated with tk

        def __play_click(self,event=None):
                #print '__play_click()'
                if self.__tabs != None:
                        if len(self.__tabs) != 0:
                                active_tab = self.__get_current_selected_tab()
                                if str(active_tab[1]) in self.__texts:
                                        pargs = ['java','-jar',str(self.__defaults['plantumlDirectory']),str(self.__texts[str(active_tab[1])][2])]
                                        #print pargs
                                        try:
                                                result = call(pargs)
                                                #print 'Success'
                                                #show Image in Toplevel Window
                                                #print self.__texts[str(active_tab[1])][2]
                                                d = str(self.__texts[str(active_tab[1])][2])[:str(self.__texts[str(active_tab[1])][2]).rindex('/')+1]
                                                f = str(self.__texts[str(active_tab[1])][2])[str(self.__texts[str(active_tab[1])][2]).rindex('/')+1:]
                                                f = f[:f.rindex('.')+1]+'png'
                                                #print d+'\t'+f
                                                umlImage = Image.open(d+f)
                                                umlPhoto = ImageTk.PhotoImage(umlImage)
                                                topLevel = tk.Toplevel()
                                                topLevel.title(f)
                                                imgLabel = tk.Label(topLevel,image=umlPhoto)
                                                imgLabel.image = umlPhoto
                                                imgLabel.pack(side=tk.TOP)
                                                textLabel = tk.Label(topLevel,text=f,font=('Helvetica',20,'bold'))
                                                textLabel.pack(side=tk.BOTTOM)
                                        except CalledProcessError, e:
                                                tkMessageBox.showinfo("UML generation failed:","Error: %s"%e)
        
        def __preferences_click(self):
                pass
        
        def __about_click(self):
                tkMessageBox.showinfo("About "+self.__defaults['mainTitle'],"Author:\tMelf Pahlke\n\nit uses:\n\nPython 2.7.9:\t www.python.org\nplantuml:\t\twww.plantuml.com\n(graphviz:\t\twww.graphviz.org)\njava:\t\twww.java.com")

        def __key_pressed_in_text(self,event=None):
                #print '__key_pressed_in_text()'
                #enable redo
                if str(event.keycode) == str(116): #F5
                        self.__play_click()
                        return
                self.__main_menu_commands['Edit'].entryconfig(0,state="normal")
                #add to undoHistory
                self.__add_to_undo_history(event=event,w=event.widget)

        def __key_released_in_text(self,event=None):
                #print str(event.keycode)+' released in '+str(event.widget)
                if str(event.keycode) == str(116): #F5
                        self.__play_click()
                        return
                if self.__compare_to_saved_file(event=event,w=event.widget) != 0:
                        #print 'differs from saved file'
                        self.__add_to_undo_history(event=event,w=event.widget)

        def __changed_size(self,event=None):
                if 'rememberLastSize' in self.__defaults:
                        if self.__defaults['rememberLastSize'] == True:
                                self.__defaults['mainWidth'] = event.width
                                self.__defaults['mainHeight'] = event.height
                                #write defaults to file
                                with open('config','w') as f:
                                        f.write(str(self.__defaults))
                                #print 'saved new size hxw [px]:\t'+str(event.height)+'x'+str(event.width)

        def __add_to_undo_history(self,event=None,w=None):
                #print '__add_to_undo_history()'
                if 'undoHistorySize' in self.__defaults:
                        maxSize = int(self.__defaults['undoHistorySize'])
                else:
                        maxSize = 10 #default
                tabName = ''
                for wdgt in self.__texts:
                        if self.__texts[wdgt][0] == w:
                                tabName = wdgt
                #print tabName
                if not '*' in tabName:
                        self.__mark_as_unsaved(tabName)
                        tabName = tabName+'*'
                if self.__undoHistory == None:
                       self.__undoHistory = {}
                if not tabName in self.__undoHistory:
                        self.__undoHistory[tabName] =[]
                        self.__main_menu_commands['Edit'].entryconfig(0,state="normal")
                if len(self.__undoHistory[tabName]) > 0:
                        old = self.__undoHistory[tabName][len(self.__undoHistory[tabName])-1]
                        new = w.get(1.0,tk.END)
                        if len(old) >= len(new):
                                for i in range(0,len(new)):
                                        if old[i] != new[i]:
                                                self.__undoHistory[tabName].append(w.get(1.0,tk.END))
                                                return
                        else:
                                for i in range(0,len(old)):
                                        if old[i] != new[i]:
                                                self.__undoHistory[tabName].append(w.get(1.0,tk.END))
                                                return
                else:
                        self.__undoHistory[tabName].append(w.get(1.0,tk.END))
                        return 0

        def __compare_to_saved_file(self,event=None,w=None):
                #print 'event:\t'+str(event)+'\nw:\t'+str(w)+'\n'
                currentContent = w.get(1.0,tk.END) #get the whole content of the selected Text widget
                #currentContent = currentContent.split('\n')
                #currentContent = currentContent[:len(currentContent)-1]
                #for el in currentContent:
                #        el = el+'\n'
                currentContent = currentContent.encode('utf-8')
                savedContent = ""
                fileLocation = ''
                for text in self.__texts:
                        #print text
                        if self.__texts[text][0] == w:
                                fileLocation = self.__texts[text][2]
                #print fileLocation
                if isfile(fileLocation):
                        with open(fileLocation,'r') as f:
                                #savedContent = f.readlines()
                                savedContent = f.read()
                #print currentContent
                #print savedContent
                for i in range(0,len(savedContent)):
                        if currentContent[i] != savedContent[i]:
                                #print currentContent[i]+'!='+savedContent[i]
                                return -1
                return 0
                #if len(currentContent) != len(savedContent):
                #        #print str(len(currentContent))+'!='+str(len(savedContent))
                #        return 1 # difference in lines
                #else:
                #        #len() equal -> compare substrings
                #        for i in range(0,len(currentContent)):
                #                #if len(currentContent[i]) != len(savedContent[i]):
                #                #        #print "line length does not match"
                #                #        #print currentContent[i]
                #                #        #print savedContent[i]
                #                #        return 2 #difference in line length
                #                for j in range(0,len(currentContent[i])):
                #                        if currentContent[i][j].encode('utf-8') != savedContent[i][j]:
                #                                #print currentContent[i][j].encode('utf-8')+'!='+savedContent[i][j]
                #                                return 3 #difference in letter
                #        return 0 # no difference found
                
        def __mark_as_unsaved(self,name,event=None,w=None):
                #print '__mark_as_unsaved('+str(name)+')'
                try:
                        i = str(name).index('*')
                        return 0
                except ValueError:
                        #print 'name: '+str(name)+' nicht markiert'
                        if name in self.__tabs and name in self.__texts:
                                tabBuf = self.__tabs[name]
                                textBuf = self.__texts[name]
                                del self.__tabs[name]
                                del self.__texts[name]
                                neu = name+'*'
                                self.__tabs[neu] = tabBuf
                                self.__texts[neu] = textBuf
                                self.__nb.tab(self.__nb.select(),text=neu)

        def __mark_as_saved(self,name,event=None,w=None):
                try:
                        if '*' in name:
                                tabBuf = None
                                textBuf = []
                                if name in self.__tabs:
                                        tabBuf = self.__tabs[name]
                                        del self.__tabs[name]
                                if name in self.__texts:
                                        textBuf = self.__texts[name]
                                        del self.__texts[name]
                                neu = name.replace('*','')
                                self.__tabs[neu] = tabBuf
                                self.__texts[neu] = textBuf
                                self.__nb.tab(self.__nb.select(),text=neu)
                                return 0
                        else:
                                return 0
                except ValueError:
                        return -1

def load_defaults():
        try:
                if isfile('config'):
                        result = ""
                        with open('config','r') as f:
                                result = literal_eval(f.read())
                        if not 'plantumlDirectory' in result:
                                #get file address
                                options = {}
                                options['defaultextension'] = '.jar'
                                options['filetypes'] = [('all files', '.*'), ('Java applications', '.jar')]
                                options['initialdir'] = "C://"
                                result['plantumlDirectory'] = tkFileDialog.askopenfilename(initialdir=options)
                                with open('config','w') as f:
                                        f.write(str(result))
                        return result
                else:
                        #create not existent file
                        defaults = {"mainWidth":800,"mainHeight":600,"mainXOffset":200,"mainYOffset":200,"mainTitle":"plantuml-tool","workingDirectory":"C:/"}
                        with open('config','a') as f:
                                f.write(str(defaults))
                        return defaults
        except IOError:
                return None
                
def main():
        dflts = load_defaults()
        if dflts != None:
                MainFrame(dflts)
                
if __name__=='__main__':
        main()
